
'******************************************************************
'     Autor: Alex D.  (allu), 1.Okt.2013
'
'     AM2302 abfragen, Temperatur und Feuchte laden
'     Der Quarztakt muss zwischen 2Mhz bis 16Mhz liegen
'    ---------------------------------------------
'
'     Programmlaufzeit: ca 5-6 msec, abhngig vom AM2302-Dateninhalt
'     Whrend der Programmlaufzeit ist kein Interupt zulssig !
'
'     Aufruf: Gosub Am2302_lesen
'
'     Rckgabewerte:
'     AM_err: 0 = Abfrage erfolgreich, Werte x 10 in:
'                                      Temperatur (integer) und
'                                      Feuchte (integer)
'             1 = Parityfehler  (Antwort fehlerhaft)
'             2 = Timerberlauf (kein AM2302 angeschlossen ? )
'******************************************************************


'------------------------------------------------------------------*
'   Konfiguration der Mikrocontrollerumgebung
'------------------------------------------------------------------*
$regfile = "m1284pdef.dat"              '_Chip = 103      'getestete AVRs
'$regfile = "m644def.dat"                '_Chip =  37

$crystal = 12800000
'2000000    '3686400    '7372800    '12800000   '1600000  'getestete Frequenzen

$hwstack = 256
$swstack = 128
$framesize = 128

'******************************************************************
'   Portzuordnung, Anschluss des AM2302 an Port_am und Pin_am
'******************************************************************
Port_in_am Alias Pina
Port_out_am Alias Porta
Pin_am Alias 0

'------------------------------------------------------------------
'  Speicher zuordnen
'------------------------------------------------------------------
Dim Am(5) As Byte                       'AM2302 Telegrammspeicher
Dim Am_parity As Byte At Am(1) Overlay
Dim Am_temperatur As Integer At Am(2) Overlay
Dim Am_feuchte As Integer At Am(4) Overlay

Dim Am_err As Byte                      '0=ok, 1=Parityfehler, 2=Timeout

'------------------------------------------------------------------
'   Register fr ASM
'------------------------------------------------------------------                                                                          '------------------------------------------------------------------
Akku_a Alias R0
Akku_b Alias R1
Akku_c Alias R2
Akku_d Alias R3
Vom_am Alias R4

Signal Alias R16
Impuls Alias R17
Notaus Alias R18
Dauer Alias R19
Bitzahl Alias R20
Zwischen Alias R21

'------------------------------------------------------------------
'   Konstanten zum Ausmessen des AM2302-Telegramms
'   Mssen bei nderungen im Assemblerprogramm neu angepast werden
'------------------------------------------------------------------
Const _timeout = _xtal / 200000
Const _schwelle =((_xtal / - 370000))

'------------------------------------------------------------------
'   Konstanten zur Fehlerbeschreibung
'------------------------------------------------------------------
Const _parity_ok = 0
Const _parity_err = 1
Const _timeout_err = 2

'------------------------------------------------------------------
'   Konstanten zur Filtereinstellung
'   Filterlnge in Abhnigkeit der Quarzfrequenz anpassen
'------------------------------------------------------------------
#if _xtal < 4000000
'  Filterlnge verkrzen
Const _am_filterlaenge = &B0000_0111
Const _am_pos_flanke = &B0000_0001
Const _am_neg_flanke = &B0000_0110
#else
'  Filterlnge ber 4Mhz
Const _am_filterlaenge = &B0000_1111
Const _am_pos_flanke = &B0000_0001
Const _am_neg_flanke = &B0000_1110
#endif


Goto Am2302_100                         'Zum AM2302-Testen, Werte anzeigen

'------------------------------------------------------------------

'******************************************************************
'     AM2302 abfragen, Temperatur und Feuchte laden
'
'******************************************************************

Am2302_lesen:

'***  1ms-Startimpuls senden
   Config Port_out_am.pin_am = Output
      Reset Port_out_am.pin_am
      Waitus 1000
   Config Port_out_am.pin_am = Input
      Waitus 3
$asm

'***  Sensordaten einlesen
Am2302:
   ldi   Signal, &B1111_1111     ;zuerst wird eine 1 anstehen, vorbesetzen
   clr   Bitzahl                 ;Bitzaehler auf 0
   ldi   Notaus, _Timeout        ;Timer Notausstieg aufziehen

Am2302_10:
   in    Zwischen, port_in_am    ;Signal vom AM2302 einlesen
   lsl   Signal                  ;... und nach links schieben
   sbrc  Zwischen, Pin_am
   inc   Signal                  ;aktueller Pegel nach D0
   andi  Signal, _am_filterlaenge;Filterlnge gegen Spikes einstellen

   cpi   Signal, _am_pos_flanke  ;Positive Flanke, mit Entprellung
   brne  Am2302_20

   LDi   Impuls, 1               ;Einsen-Zhler "Dauer" einschalten
   ldi   Dauer, _Schwelle        ;"Dauer" mit der _Schwelle laden
   ldi   Notaus, _Timeout        ;Timer Notausstieg aufziehen
   rjmp  Am2302_30

Am2302_20:
   cpi   Signal, _am_neg_flanke  ;Negative Flanke, mit Entprellung
   brne  Am2302_30
                                 ;Jetzt ist minimal 45sec Zeit zum rechnen
   inc   Bitzahl                 ;Bit zu Ende, Bitzhler  erhhen
   LDi   Impuls, 0               ;Einsen-Zhler "Dauer" ausschalten
   ldi   Notaus, _Timeout        ;Timer Notausstieg aufziehen

  ;*** Ergebnis einrollen
   Lds   Zwischen, {am(1)}              'Parity
   lsl   Zwischen
   sbrs  Dauer, 7                ;Negativ = 0 , Positiv = 1
   inc   Zwischen
   sts   {am(1)}, Zwischen

   Lds   Zwischen, {am(2)}              'Temperatur low
   rol   Zwischen
   sts   {am(2)}, Zwischen

   Lds   Zwischen, {am(3)}              'Temperatur high
   rol   Zwischen
   sts   {am(3)}, Zwischen

   Lds   Zwischen, {am(4)}              'Feuchte low
   rol   Zwischen
   sts   {am(4)}, Zwischen

   Lds   Zwischen, {am(5)}              'Feuchte high
   rol   Zwischen
   sts   {am(5)}, Zwischen

Am2302_30:
   sbrc  Impuls, 0               ;Solange eine 1 anliegt
   Inc   Dauer                   ; ...den Bitzhler erhhen

   cpi   Bitzahl, 42             ;Alle Bits eingelesen ?
   breq  AM2302_40

   dec   Notaus                  ;Timer Notausstieg runterzhlen
   breq  AM2302_50               ;Zeitdauer-berlauf prfen

   rjmp  Am2302_10               ;Noch nicht alle Bits da, also weitermachen

Am2302_40:
   Lds   Vom_am, {am(1)}                'Parity vom AM2302
   Lds   Akku_a, {am(2)}                'Temperatur low
   Lds   Akku_b, {am(3)}                'Temperatur high
   Lds   Akku_c, {am(4)}                'Feuchte low
   Lds   Akku_d, {am(5)}                'Feuchte high

   add   Akku_a, Akku_b
   add   Akku_a, Akku_c
   add   Akku_a, Akku_d

   Ldi   Zwischen, _parity_ok
   cp    Akku_a, Vom_am                   ;Parityprfung
   Breq  Am2302_99

   Ldi   Zwischen, _parity_err
   rjmp  Am2302_99

Am2302_50:
   Ldi   Zwischen, _timeout_err

Am2302_99:
   sts   {AM_err}, Zwischen               ;Fehlerkennung eintragen
   ret

$end Asm


'********************************************************************
'   Anfang Testprogramm zur Anzeige der AM2302-Daten
'********************************************************************
Am2302_100:

Dim Am_err_parity As Word , Am_err_timeout As Word
Dim Temp As String * 5 , Relf As String * 5

'*** Display an die verwendete Hardware anpassen:
Config Lcdpin = Pin , Db4 = Porta.2 , Db5 = Porta.5 , Db6 = Porta.3 , Db7 = Porta.4 , E = Porta.6 , Rs = Porta.7       ' , Mode = 4
Config Lcd = 16x2

Cls
Locate 1 , 1
Lcd "  AM2302 alle"
Locate 2 , 1
Lcd " 3sec auslesen !"

Do
   Wait 3                               'Nur weil es ein Testprogramm ist,
                                          ' ... sonst nie.

   Gosub Am2302_lesen                   'In 6msec bin ich wieder da !


   If Am_err = 1 Then Incr Am_err_parity
   If Am_err = 2 Then Incr Am_err_timeout

   Temp = Str(am_temperatur)
   Temp = Format(temp , "0.0")
   Relf = Str(am_feuchte)
   Relf = Format(relf , "0.0")

   'Anzeige der Temperatur in Grad
   '... der Feuchte in %
   Locate 1 , 1
   Lcd Temp ; Chr(&B1101_1111) ; "C " ; Relf ; "rF%      "

   'Zeigt an, ob ein Fehler aufgetreten ist und wieviel
   '... Parity- und Timeoutfehler seit Programmstart aufgetreten sind.
   Locate 2 , 1
   Lcd ; "E=" ; Str(am_err) ; " P=" ; Str(am_err_parity) ; " T=" ; Str(am_err_timeout) ; "       "

Loop

'********************************************************************
' Ende Testprogramm
'********************************************************************
End